"use strict";
//    Copyright 2018 Quirino Brizi [quirino.brizi@gmail.com]
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
Object.defineProperty(exports, "__esModule", { value: true });
exports.RouteHandler = void 0;
const dino_core_1 = require("dino-core");
const request_body_parser_1 = require("./request/request.body.parser");
const interface_1 = require("./interface");
const helper_1 = require("./helper");
const http_exception_1 = require("./exception/http.exception");
// const DEFAULT_RESPONSE = { content: { 'application/json': { schema: { type: 'object', properties: { message: { type: 'string' } } } } } };
/**
 * Handles the route subject to the client request, it does act as a express middleware.
 *
 */
class RouteHandler {
    constructor(applicationContext, api, responseValidator, environment) {
        this.api = api;
        this.applicationContext = applicationContext;
        this.responseValidator = responseValidator;
        this.environment = environment;
        this.requestBodyParser = new request_body_parser_1.RequestBodyParser(environment);
    }
    /**
     * Handle the request from the client forwarding the request to the Interface that implement the
     * method defined on the API operation identifier.
     *
     * @param {any} req the express request object
     * @param {any} res the express response object
     * @param {Function} next a callback to propagate the handling action to the next middleware.
     *
     * @public
     */
    handle(req, res, next) {
        try {
            this.ensureInterface();
            const answer = this._doHandleRequest(req);
            res.body = answer;
            // this.handleApiResponse(answer, res, next);
        }
        catch (e) {
            dino_core_1.Logger.error(`Unable to complete transaction ${e}`);
            this.handleError(e, res, next);
        }
        next();
    }
    _doHandleRequest(req) {
        const apiHandlerType = helper_1.Helper.getVariable(this.environment, 'dino:api:handler:type', 'proxy');
        let answer;
        if ('proxy' == apiHandlerType) {
            const params = this.collectParameters(req);
            const body = this.requestBodyParser.parse(req);
            params['req'] = req;
            params['body'] = body;
            answer = this._interface[this.api.operationId](params);
        }
        else {
            // propagate the entire express request
            answer = this._interface[this.api.operationId](req);
        }
        return answer;
    }
    /**
     * Collect the parameters defined as part of the OpenAPI definition.
     * @param {Object} req the express request
     */
    collectParameters(req) {
        const parameters = helper_1.Helper.ensureValue(this.api.parameters, []);
        if (!parameters.reduce) {
            return [];
        }
        return parameters.reduce((acc, param) => {
            const name = param.name.replace(/-/g, '_');
            switch (param.in) {
                case 'path':
                    acc[name] = helper_1.Helper.ensureValue(req.params[param.name], param.schema.default);
                    break;
                case 'query':
                    acc[name] = helper_1.Helper.ensureValue(req.query[param.name], param.schema.default);
                    break;
                case 'header':
                    acc[name] = helper_1.Helper.ensureValue(req.headers[param.name], param.schema.default);
                    break;
            }
            return acc;
        }, {});
    }
    handleApiResponse(answer, res, next) {
        if (helper_1.Helper.isAPromise(answer)) {
            return answer.then((response) => this.sendResponse(response, res, next)).catch((error) => this.handleError(error, res, next));
        }
        else {
            return this.sendResponse(answer, res, next);
        }
    }
    handleError(error, _response, next) {
        var _a;
        // response.serviceErrorResponse = error;
        // Logger.error('unable to complete the requested transaction', error);
        // const statusCode = error.statusCode || error.status || 500;
        // const definedResponse = this.api.responses[statusCode] || this.api.responses['default'] || DEFAULT_RESPONSE;
        // const contentType = Object.keys(definedResponse.content)[0];
        // if (!error.code) {
        //   error.code = statusCode;
        // }
        // clientResponse
        //   .status(statusCode)
        //   .type(contentType)
        //   .send(this._buildClientErrorResponse(definedResponse.content[contentType].schema.properties || {}, contentType, error));
        let err;
        if (!dino_core_1.ObjectHelper.instanceOf(error, http_exception_1.HttpException)) {
            err = http_exception_1.HttpException.create(500, (_a = error === null || error === void 0 ? void 0 : error.message) !== null && _a !== void 0 ? _a : 'Unknown error');
        }
        next(err);
    }
    sendResponse(answer, response, next) {
        response.body = answer;
        if (this.responseValidator) {
            try {
                this.responseValidator(response);
            }
            catch (e) {
                this.handleError(e, response, next);
            }
        }
        // const responseStatusCode = Number(answer.statusCode || answer.status);
        // if (responseStatusCode >= 400) {
        //   return this.handleError(answer, response, next);
        // }
        // // const response = this.__normaliseAnswer(answer);
        // // will throw ValidationError if validation fails, this is handled on the
        // // handleError method
        const statusCode = Object.keys(this.api.responses || {}).find((sc) => sc.startsWith('2')) || '200';
        response.status(parseInt(statusCode, 10));
        // if (this.responseValidator) {
        //   this.responseValidator(response);
        // }
        // Handle response not of type Response
        // response.propagate(res);
        next();
    }
    /**
     * Normalise the response from the API to an instance of Response
     * @param {Any} answer the response from the API
     * @returns {Response}
     *
     * @private
     */
    // __normaliseAnswer(answer) {
    //   let response = !answer ? Response.build({}) : answer;
    //   if (!Helper.instanceOf(response, Response)) {
    //     response = Response.build(answer);
    //   }
    //   return response;
    // }
    // _buildClientErrorResponse(properties, contentType, error) {
    //   return Helper.format(
    //     Object.keys(properties).reduce((resp, prop) => {
    //       resp[prop] = resp[prop] = error[prop] || (error['body'] || {})[prop] || '';
    //       return resp;
    //     }, {}),
    //     contentType,
    //   );
    // }
    /**
     * Lookup the interface that can handle the defined operationId for the API.
     * @returns {Interface}
     *
     * @private
     */
    ensureInterface() {
        if (!this._interface) {
            const interfaces = this.applicationContext.resolveAll(interface_1.Interface);
            this._interface = (interfaces || []).find((_interface) => {
                try {
                    return Reflect.has(_interface, this.api.operationId);
                }
                catch (e) {
                    return false;
                }
            });
            if (!this._interface) {
                throw Error(`there are no interface defined with operationId ${this.api.operationId}`);
            }
        }
    }
    /**
     * Define tge scope as transient so that a new instance will be created for every request
     *
     * @static
     * @public
     */
    static scope() {
        return 'transient';
    }
}
exports.RouteHandler = RouteHandler;
//# sourceMappingURL=route.handler.js.map